# Copyright 2013-2014 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

from contextlib import contextmanager

from mock import (
    create_autospec,
    patch,
)
from pyelasticsearch import ElasticSearch

from charmworld.testing import logger_buffer
from charmworld.jobs.config import (
    CHARM_QUEUE,
    BASKET_QUEUE,
    settings,
)
from charmworld.jobs.worker import (
    QueueWorker,
    main,
)
from charmworld.jobs.ingest import (
    DBIngestJob,
    UpdateCharmJob,
)
from charmworld.jobs.tests.test_ingest import charm_update_environment
from charmworld.jobs.utils import get_queue
from charmworld.search import ElasticSearchClient
from charmworld.testing import (
    factory,
    MongoTestBase,
)


class SuccessfulRun(DBIngestJob):

    def run(self, charm_data):
        charm_data['ingested'] = True


class FailureRun(DBIngestJob):

    def run(self, charm_data):
        if charm_data['charm'] == 'foo':
            raise Exception('Exception forced by job.')
        else:
            charm_data['ingested'] = True


class GatheringRun(DBIngestJob):

    def setup(self, db=None):
        super(GatheringRun, self).setup(db)
        self.charms = []

    def run(self, charm_data):
        self.charms.append(charm_data)


class FakeArgs(object):
    run_forever = False
    debug = False


@contextmanager
def main_environment(charm, index_client, queue):
    """An environment in which worker.main() can be run."""
    with patch('charmworld.jobs.worker.configure_logging'):
        with charm_update_environment(charm, index_client):
            yield


def FauxWorkerFactory(test_case, expected_arg_hash):
    class FauxWorker:
        run_called = False

        def __init__(self, *ignored):
            pass

        @classmethod
        def run(cls, args_):
            for argname, expected in expected_arg_hash.items():
                actual = args_.__getattribute__(argname)
                test_case.assertEqual(expected, actual)
            cls.run_called = True

    return FauxWorker


class WorkerTest(MongoTestBase):

    def setUp(self):
        super(WorkerTest, self).setUp()
        self.charm_queue = get_queue(self.db, CHARM_QUEUE)
        self.basket_queue = get_queue(self.db, BASKET_QUEUE)
        self.charm_queue.put({'charm': 'foo'})
        self.charm_queue.put({'charm': 'bar'})
        self.args = FakeArgs()

    def tearDown(self):
        self.charm_queue.clear()
        super(WorkerTest, self).tearDown()

    def test_basic_running(self):
        # Barring errors, the worker will go through its entire set of
        # ingest jobs for a charm.
        runners = (
            SuccessfulRun(),
            GatheringRun(),
        )
        ingest_jobs = (
            (self.charm_queue, runners),
        )
        worker = QueueWorker(self.db, None, ingest_jobs)

        worker.run(self.args)

        self.assertEqual(0, self.charm_queue.size())
        expected = [
            {'ingested': True, 'charm': 'foo'},
            {'ingested': True, 'charm': 'bar'},
        ]
        charms = runners[1].charms
        self.assertEqual(expected, charms)

    def test_failure_while_running(self):
        runners = (FailureRun(), GatheringRun())
        ingest_jobs = (
            (self.charm_queue, runners),
        )
        worker = QueueWorker(self.db, queue_jobs=None,
                             ingest_jobs=ingest_jobs)
        worker.run(self.args)

        self.assertEqual(0, self.charm_queue.size())
        expected = [
            {'ingested': True, 'charm': 'bar'},
        ]
        charms = runners[1].charms
        self.assertEqual(expected, charms)

    def test_stop_when_next_item_none(self):

        class LockedJobQueue:
            """Queue representing the state when a locked jobs exists."""

            @staticmethod
            def next():
                """A locked job cannot be processed."""
                return None

            @staticmethod
            def size():
                """If a queue has a locked job, its length > 0"""
                return 1

        jobs = (
            SuccessfulRun(),
            GatheringRun(),
        )
        ingest_jobs = (
            (LockedJobQueue, jobs),
        )
        worker = QueueWorker(self.db, None, ingest_jobs)
        worker.run(self.args)

    def test_ingest_job_contract(self):
        # All ingest jobs accept the right inputs.
        ingest_jobs = (
            (self.charm_queue, [create_autospec(UpdateCharmJob)()]),
        )
        worker = QueueWorker(self.db, None, ingest_jobs)
        worker.run(self.args)
        self.assertEqual(0, self.charm_queue.size())

    def test_interval_is_int(self):
        ingest_jobs = (
            (self.charm_queue, []),
        )
        worker = QueueWorker(self.db, None, ingest_jobs)
        self.assertIs(int, type(worker.interval),
                      'QueueWorker.interval is a %s, not an int.' %
                      type(worker.interval))

    def run_main_with_exit(self, extra_args=None):
        with patch('charmworld.jobs.worker.configure_logging'):
            with self.assertRaises(SystemExit) as e:
                fake_argv = ['ingest-queued']
                if extra_args is not None:
                    # Listify and add.
                    fake_argv.extend(extra_args.split())
                main(fake_argv, db=self.db)
        return e.exception.code

    def test_elasticsearch_failure(self):
        payload = factory.get_payload_json()
        factory.add_store_data(payload)
        self.charm_queue.clear()
        self.charm_queue.put(payload)
        charm = factory.get_charm_json(payload=payload)
        index_client = ElasticSearchClient(
            ElasticSearch('http://localhost:70'), 'foo')
        with patch.dict(settings,
                        values={'es_urls': 'http://localhost:70'}):
            with main_environment(charm, index_client, self.charm_queue):
                with logger_buffer('charm.worker') as handler:
                    with patch('charmworld.jobs.worker.lp.run'):
                        exit_code = self.run_main_with_exit()
        self.assertEqual(40, handler.buffer[1].levelno)
        self.assertIn('Could not connect', handler.buffer[1].msg)
        self.assertEqual(1, exit_code)
        self.assertEqual(0, self.charm_queue.size())

    def test_job_for_bundles_is_run_if_bundle_queue_has_entries(self):
        # Basket queue entries are processed by run_job.
        queue = get_queue(self.db, BASKET_QUEUE)
        queue_entry = {'bundle': 'foo'}
        queue.put(queue_entry)
        with patch('charmworld.jobs.worker.lp.run'):
            with patch('charmworld.jobs.worker.run_job') as run_job:
                self.run_main_with_exit()
                queue_entries_run = [x[0][1] for x in run_job.call_args_list]
                self.assertIn(queue_entry, queue_entries_run)

    def test_worker_does_not_run_forever_by_default(self):
        expected = {'run_forever': False}
        FauxWorker = FauxWorkerFactory(self, expected)
        with patch('charmworld.jobs.worker.QueueWorker', new=FauxWorker):
            self.run_main_with_exit()
        self.assertTrue(FauxWorker.run_called)

    def test_worker_runs_forever_if_command_line_says_to(self):
        expected = {'run_forever': True}
        FauxWorker = FauxWorkerFactory(self, expected)
        with patch('charmworld.jobs.worker.QueueWorker', new=FauxWorker):
            self.run_main_with_exit('--run-forever')
        self.assertTrue(FauxWorker.run_called)

    def test_lp_run_limit(self):
        # Show that the '--limit' argument is passed and used.
        expected = {'limit': 5}
        FauxWorker = FauxWorkerFactory(self, expected)
        with patch('charmworld.jobs.worker.QueueWorker', new=FauxWorker):
            self.run_main_with_exit('--limit 5')
        self.assertTrue(FauxWorker.run_called)

    def test_lp_run_debug_set(self):
        # Show that the '--debug' argument is passed and used.
        expected = {'debug': True}
        FauxWorker = FauxWorkerFactory(self, expected)
        with patch('charmworld.jobs.worker.QueueWorker', new=FauxWorker):
            self.run_main_with_exit('--debug')
        self.assertTrue(FauxWorker.run_called)

    def test_lp_run_debug_not_set(self):
        # Show that the '--debug' argument is False by default.
        expected = {'debug': False}
        FauxWorker = FauxWorkerFactory(self, expected)
        with patch('charmworld.jobs.worker.QueueWorker', new=FauxWorker):
            self.run_main_with_exit()
        self.assertTrue(FauxWorker.run_called)

    def test_lp_run_prefix(self):
        # Show that the '--prefix' argument is passed and used.
        expected = {'prefix': 'abentley'}
        FauxWorker = FauxWorkerFactory(self, expected)
        with patch('charmworld.jobs.worker.QueueWorker', new=FauxWorker):
            self.run_main_with_exit('--prefix abentley')
        self.assertTrue(FauxWorker.run_called)

    def test_lp_run_clear(self):
        # Show that the '--prefix' argument is passed and used.
        expected = {'clear': True}
        FauxWorker = FauxWorkerFactory(self, expected)
        with patch('charmworld.jobs.worker.QueueWorker', new=FauxWorker):
            self.run_main_with_exit('--clear')
        self.assertTrue(FauxWorker.run_called)

    def test_wait_seconds_time_remaining(self):
        # The waiting time is computed correctly if there is still time
        # remaining.
        interval = 1000
        start = 10000
        now = start + interval / 2
        expected = interval - (now - start)
        self.assertEqual(
            expected, QueueWorker.wait_seconds(interval, start, now))

    def test_wait_seconds_no_time_remaining(self):
        # The waiting time is properly set to zero if there is not any
        # remaining time.
        interval = 1000
        start = 10000
        now = start + interval + 5
        self.assertEqual(
            0, QueueWorker.wait_seconds(interval, start, now))
