# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import argparse
import logging
import sys
import time

from charmworld.jobs.config import (
    CHARM_QUEUE,
    BASKET_QUEUE,
)
from charmworld.jobs.ingest import (
    run_job,
    UpdateBundleJob,
    UpdateCharmJob,
)
from charmworld.jobs import lp
from charmworld.jobs.utils import get_queue
from charmworld.models import getconnection
from charmworld.models import getdb
from charmworld.search import SearchServiceNotAvailable
from charmworld.utils import (
    configure_logging,
    get_ini,
)


settings = get_ini()
INTERVAL = int(settings.get('worker_interval'))


class QueueWorker(object):
    """Processes raw charm data"""

    def __init__(self, db, queue_jobs, ingest_jobs, interval=INTERVAL):
        self.db = db
        self.queue_jobs = queue_jobs or []
        self.ingest_jobs = ingest_jobs or []
        self.log = logging.getLogger('charm.worker')
        self.interval = interval

    @staticmethod
    def wait_seconds(interval, start, now):
        remaining = interval - (now - start)
        return int(max(0, remaining))

    def run(self, args, now=None):
        if args.debug:
            self.log.setLevel(logging.DEBUG)
        self.log.info(
            'Starting processing with interval {} seconds.'.format(
                self.interval))
        self.start_time = now if now is not None else time.time()
        return_code = 0
        for _, jobs in self.ingest_jobs:
            for job in jobs:
                job.setup(db=self.db)
        while True:
            for job in self.queue_jobs:
                job(args)
            # The clear argument is only used on the first queueing run.
            # Unset it so it subequent passes will not clear the queue.
            args.clear = False
            for queue, jobs in self.ingest_jobs:
                while True:
                    item = queue.next()
                    if item is None:
                        break
                    try:
                        for job in jobs:
                            try:
                                if not run_job(
                                        job, item.payload, needs_setup=False):
                                    break
                            except SearchServiceNotAvailable as e:
                                self.log.exception(str(e))
                                queue.clear()
                                return_code = 1
                    finally:
                        item.complete()

            if not args.run_forever:
                self.log.info('Empty queue and not running forever; quitting.')
                break

            wait_seconds = self.wait_seconds(
                self.interval, self.start_time, time.time())
            self.log.info(
                'Empty queue.  Waiting %s seconds to recheck.' %
                wait_seconds)
            time.sleep(wait_seconds)
        return return_code


def main(argv=sys.argv, db=None):
    if db is None:
        settings = get_ini()
        connection = getconnection(settings)
        db = getdb(connection, settings.get('mongo.database'))
    configure_logging()
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--run-forever',
        help='Keep running after the queues are processed and process more '
        'later.',
        action='store_true', dest='run_forever')
    lp.configure_parser(parser)
    # Explicitly pass sys.argv to make monkeypatching more reliable.
    args = parser.parse_args(argv[1:])
    queue_jobs = (lp.run, )
    ingest_jobs = (
        (get_queue(db, CHARM_QUEUE), [UpdateCharmJob(args.debug)]),
        (get_queue(db, BASKET_QUEUE), [UpdateBundleJob(args.debug)]),
    )
    worker = QueueWorker(db, queue_jobs, ingest_jobs)
    sys.exit(worker.run(args))
