# Copyright 2012, 2013, Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE.txt).

from cStringIO import StringIO
import os

from charmworld.models import (
    Charm,
    CharmSource,
    construct_charm_id,
    FeaturedSource,
)
from charmworld.scripts import remove_charm
from charmworld.testing import MongoTestBase
from charmworld.testing.factory import make_charm
from charmworld.utils import project_root


class RemoveCharmScriptTests(MongoTestBase):

    def setUp(self):
        super(RemoveCharmScriptTests, self).setUp()
        # This is a lot of work just to create a new revision of a charm.
        older_charm = make_charm(self.db)
        newer_charm_data = older_charm._representation
        newer_charm_data['store_data']['revision'] += 1
        newer_charm_data['_id'] = construct_charm_id(newer_charm_data)
        self.db.charms.insert(newer_charm_data)
        self.charm = Charm(newer_charm_data)
        # As a convenience we will create a stderr stand-in here.
        self.stderr = StringIO()
        # Be paranoid about the charms actually existing in the database.
        assert self.db.charms.find_one({'_id': older_charm._id})
        assert self.db.charms.find_one({'_id': self.charm._id})

    def is_indexed(self, id_):
        return self.index_client.get(id_) is not None

    def test_entry_point_exists(self):
        # The entry point is wrapped in a script by setuptools.  We want to be
        # sure the entrypoint actually exists.
        self.assertIn(
            'remove-charm = charmworld.scripts.remove_charm:entry_point\n',
            open(os.path.join(project_root, 'setup.py')).read())

    def test_script_chides_you_for_not_providing_an_id(self):
        # If no charm ID is provided on the command line the script prints a
        # helpful message.
        exit_status = remove_charm.main([], self.stderr)
        self.assertEqual(
            self.stderr.getvalue(),
            'Please provide the ID of the charm you wish to remove.\n')
        # A non-zero exit status is set to indicate the error condition.
        self.assertGreater(exit_status, 0)

    def test_script_reports_unknown_charm_ids(self):
        # The script will generate an error message if the given charm ID does
        # not exist in the DB.
        exit_status = remove_charm.main(['', 'not a charm ID'], self.stderr)
        self.assertIn(
            "Unknown charm ID: 'not a charm ID'\n",
            self.stderr.getvalue())
        # The user is also given a hint as to how the ID is structured.
        self.assertIn(remove_charm.CHARM_ID_MESSAGE, self.stderr.getvalue())
        # A non-zero exit status is set to indicate the error condition.
        self.assertGreater(exit_status, 0)

    def test_script_removes_charm_from_mongodb(self):
        # The script removes the charm with the provided ID from the database.
        # ...we can invoke the script to delete it.
        exit_status = remove_charm.main(['', self.charm._id], self.stderr)
        # Lo and behold, the charm is actually gone.
        self.assertFalse(self.db.charms.find_one({'_id': self.charm._id}))
        # Silence is a virtue for sucessfully executing programs.
        self.assertEqual(self.stderr.getvalue(), '')
        # The exit status signals success.
        self.assertEqual(exit_status, 0)

    def test_script_removes_charm_from_featured(self):
        # If a charm is featured, its featured status is revoked by the script.
        source = FeaturedSource.from_db(self.db)
        charm_data = self.charm._representation
        source.set_featured(charm_data, 'charm')
        exit_status = remove_charm.main(['', self.charm._id], self.stderr)
        self.assertFalse(source.is_featured(charm_data, 'charm'))
        # The exit status signals success.
        self.assertEqual(exit_status, 0)

    def test_script_removes_charm_from_elastic_search(self):
        # The charm needs to be indexed so the script can remove it.
        self.use_index_client()
        charm_data = self.charm._representation
        self.index_client.index_charm(charm_data)
        # Ensure the charm actually got indexed.
        assert self.is_indexed(charm_data['_id'])
        # Removing a charm should also remove it from the search engine.
        exit_status = remove_charm.main(
            ['', self.charm._id], self.stderr, index_client=self.index_client)
        # The script removed the charm from the index.
        self.assertFalse(self.is_indexed(charm_data['_id']))
        # The exit status signals success.
        self.assertEqual(exit_status, 0)

    def test_old_revisions_of_the_charm_are_removed_too(self):
        # Requesting the removal of a charm results in all revisions of that
        # charm being removed.
        charm_data = self.charm._representation
        # This can't be the best way to find all the revisions of a particular
        # charm.
        source = CharmSource(self.db, index_client='fake')
        # There should exist more than one revision of the charm in the
        # database.
        assert len(source.find_all_revisions(charm_data)) > 1
        # Run the script.
        exit_status = remove_charm.main(['', self.charm._id], self.stderr)
        # The script removed all the charm revisions from the database.
        self.assertEqual(len(source.find_all_revisions(charm_data)), 0)
        # The exit status signals success.
        self.assertEqual(exit_status, 0)
