from cStringIO import StringIO
import re
import os
import time

import mock

from charmworld.scripts.remove_server_start_time import (
    remove_server_start_time_entries,
)
from charmworld.testing import MongoTestBase
from charmworld.utils import (
    project_root,
    record_server_start_time,
    SERVER_STARTED,
)


class RemoveServerStartTimeScriptTests(MongoTestBase):

    def test_record_server_start_time(self):
        # record_server_start_time() records the hostname and the current
        # time in a MongoDB record.
        def make_uname(hostname):
            def uname():
                return (None, hostname)
            return uname

        with mock.patch.object(
                os, 'uname', new_callable=make_uname,
                hostname='foo'):
            before = time.time()
            record_server_start_time(self.db)
            after = time.time()
        foo_start_time = self.db.server_status.find_one(SERVER_STARTED)['foo']
        self.assertTrue(before <= foo_start_time <= after)

        # Updating a server's start time retains already recorded data.
        with mock.patch.object(
                os, 'uname', new_callable=make_uname,
                hostname='bar'):
            record_server_start_time(self.db)
        recorded = self.db.server_status.find_one(SERVER_STARTED)
        self.assertEqual(set(('_id', 'foo', 'bar')), set(recorded))

        # Existing records are updated.
        with mock.patch.object(
                os, 'uname', new_callable=make_uname,
                hostname='foo'):
            before = time.time()
            record_server_start_time(self.db)
            after = time.time()
        recorded = self.db.server_status.find_one(SERVER_STARTED)
        self.assertTrue(recorded['foo'] > foo_start_time)

    def test_remove_server_start_time_entries(self):
        # Each time a server starts it stores the current time in the DB.
        # Stale entries can be removed by calling
        # remove_server_start_time_entries() (used in the script
        # remove-start-time-entry).
        self.db.server_status.insert({
            '_id': SERVER_STARTED,
            'foo': 123.456,
            'bar': 234.567,
        })
        output = StringIO()
        self.assertEqual(
            0, remove_server_start_time_entries(['', 'foo'], output))
        start_times = self.db.server_status.find_one({'_id': SERVER_STARTED})
        self.assertEqual(
            {
                '_id': SERVER_STARTED,
                'bar': 234.567,
            },
            start_times)
        self.assertEqual('', output.getvalue())

    def test_remove_server_start_time_entries_no_args(self):
        # If the server start time script is run without arguments it
        # generates an error message.
        output = StringIO()
        self.assertEqual(
            1, remove_server_start_time_entries([''], output))
        self.assertEqual(
            'At least one server hostname must be specified.\n',
            output.getvalue())

    def test_remove_server_start_time_entries_no_db_record(self):
        # If there are no server start times yet and the script is run, it
        # informs the user.
        output = StringIO()
        self.assertEqual(
            1, remove_server_start_time_entries(['', 'foo'], output))
        self.assertEqual(
            'No server start times are yet recorded.\n', output.getvalue())

    def test_remove_server_start_time_entries_invalid_server_name(self):
        # If the script is asked to remove a server start time that does not
        # exist, it generates an informative message.
        self.db.server_status.insert({
            '_id': SERVER_STARTED,
            'foo': 123.456,
        })
        output = StringIO()
        self.assertEqual(
            1, remove_server_start_time_entries(['', 'bar'], output))
        self.assertEqual(
            'hostname bar not found.\n', output.getvalue())

    def test_entry_point_exists(self):
        # The entry point is wrapped in a script by setuptools.  We want to be
        # sure the entrypoint actually exists.
        setup_py = open(os.path.join(project_root, 'setup.py')).read()
        # Remove continuation characters so searching for the config is
        # easier.
        setup_py = re.sub(r'\\\n', ' ', setup_py)
        # Also strip out newlines and collapse multiple spaces.
        setup_py = re.sub('[\n ]+', ' ', setup_py)
        script_config = (
            'remove-start-time-entry = '
            'charmworld.scripts.remove_server_start_time:entry_point')
        self.assertIn(script_config, setup_py)
