# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from mock import patch
import os
import re

from charmworld.health import (
    check_api2,
    check_api3,
    check_bundles_collection,
    check_bzr_revision,
    check_charms_collection,
    check_collections,
    check_charm_index,
    check_crontab,
    check_elasticsearch_status,
    check_ingest_queues,
    check_last_ingest_job,
    check_server_start_time,
)
from charmworld.models import FeaturedSource
from charmworld.testing import (
    factory,
    ViewTestBase,
)
from charmworld.utils import (
    LAST_INGEST_JOB_FINISHED,
    record_server_start_time,
    set_status_timestamp,
)


class TestCheckMixin:

    def assertCheck(self, check, name, status, remark=None):
        self.assertEqual(name, check.name)
        self.assertEqual(status, check.status)
        if remark is not None:
            self.assertEqual(remark, check.remark)

    @staticmethod
    def makeFailCallable():
        def failing(*args, **kwargs):
            raise ValueError
        return failing


class HealthTestCase(ViewTestBase, TestCheckMixin):

    def test_check_charms_collection_pass(self):
        factory.makeCharm(self.db)
        check = check_charms_collection(self.getRequest())
        remark = '1 charms found'
        self.assertCheck(check, 'charms collection', 'Pass', remark)

    def test_check_charms_collection_missing_data_fail(self):
        check = check_charms_collection(self.getRequest())
        remark = 'There are no charms. Is ingest running?'
        self.assertCheck(check, 'charms collection', 'Fail', remark)

    def test_check_charms_collection_no_collection_fail(self):
        request = self.getRequest()
        with patch.object(request, 'db', new_callable=self.makeFailCallable):
            check = check_charms_collection(request)
        self.assertCheck(check, 'charms collection', 'Fail')

    def test_check_bundles_collection_pass(self):
        factory.makeBundle(self.db)
        check = check_bundles_collection(self.getRequest())
        remark = '1 bundles found'
        self.assertCheck(check, 'bundles collection', 'Pass', remark)

    def test_check_bundles_collection_missing_data_fail(self):
        check = check_bundles_collection(self.getRequest())
        remark = 'There are no bundles. Is ingest running?'
        self.assertCheck(check, 'bundles collection', 'Fail', remark)

    def test_check_bundles_collection_no_collection_fail(self):
        request = self.getRequest()
        with patch.object(request, 'db', new_callable=self.makeFailCallable):
            check = check_bundles_collection(request)
        self.assertCheck(check, 'bundles collection', 'Fail')

    def test_check_charm_index_pass(self):
        charm_data = factory.makeCharm(self.db)[1]
        self.use_index_client()
        self.index_client.index_charm(charm_data)
        check = check_charm_index(self.getRequest())
        remark = '1 matches found'
        self.assertCheck(check, 'charms index', 'Pass', remark)

    def test_check_charm_index_missing_data_fail(self):
        self.use_index_client()
        request = self.getRequest()
        check = check_charm_index(request)
        remark = 'There are no matches. Is ingest running?'
        self.assertCheck(check, 'charms index', 'Fail', remark)

    def test_check_charm_index_no_index_fail(self):
        request = self.getRequest()
        # The index is not setup.
        check = check_charm_index(request)
        self.assertCheck(check, 'charms index', 'Fail')

    def _check_api_pass(self, function, version):
        charm_data = factory.makeCharm(self.db, downloads=5)[1]
        self.use_index_client()
        self.index_client.index_charm(charm_data)
        FeaturedSource.from_db(self.db).set_featured(charm_data, 'charm')
        check = function(self.getRequest())
        remark = 'Interesting has new, popular, and featured.'
        self.assertCheck(check, 'API%s interesting' % version, 'Pass', remark)

    def _check_api_missing_data_fail(self, function, version):
        charm_data = factory.makeCharm(self.db)[1]
        self.use_index_client()
        self.index_client.index_charm(charm_data)
        # There are no featured or downloaded charms.
        check = function(self.getRequest())
        remark = (
            'Interesting is missing new (found 1), popular (found 1), or '
            'featured (found 0).')
        self.assertCheck(check, 'API%s interesting' % version, 'Fail', remark)

    def _check_api_no_matches_fail(self, function, version):
        request = self.getRequest()
        # The index is not setup for the interesting search.
        check = function(request)
        self.assertCheck(check, 'API%s interesting' % version, 'Fail')

    def test_check_api2_pass(self):
        self._check_api_pass(check_api2, 2)

    def test_check_api2_missing_data_fail(self):
        self._check_api_missing_data_fail(check_api2, 2)

    def test_check_api2_no_matches_fail(self):
        self._check_api_no_matches_fail(check_api2, 2)

    def test_check_api3_pass(self):
        self._check_api_pass(check_api3, 3)

    def test_check_api3_missing_data_fail(self):
        self._check_api_missing_data_fail(check_api3, 3)

    def test_check_api3_no_matches_fail(self):
        self._check_api_no_matches_fail(check_api3, 3)

    def test_check_collections_pass(self):
        request = self.getRequest()

        def make_all_collection_names():
            def all_collection_names():
                return [
                    'fs.chunks', 'system.indexes', 'basket-queue', 'baskets',
                    'bundles', 'charm-queue', 'charms']
            return all_collection_names

        with patch.object(request.db, 'collection_names',
                          new_callable=make_all_collection_names):
            check = check_collections(request)
        self.assertCheck(
            check, 'MongoDB collections', 'Pass',
            'All required collections exist.')

    def test_check_collections_missing_collections(self):
        request = self.getRequest()
        check = check_collections(request)
        self.assertCheck(
            check, 'MongoDB collections', 'Fail',
            'Required collections do not (yet) exist: basket-queue, baskets, '
            'bundles, charm-queue, charms')

    def test_check_collections_db_error(self):
        request = self.getRequest()
        with patch.object(request.db, 'collection_names',
                          new_callable=self.makeFailCallable):
            check = check_collections(request)
        self.assertCheck(
            check, 'MongoDB collections', 'Fail',
            "Cannot query the MongoDB's collection names.")

    def test_check_ingest_queues(self):
        request = self.getRequest()
        check = check_ingest_queues(request)
        self.assertCheck(
            check, 'Ingest queue sizes', 'Pass',
            'queued charms: 0, queued baskets: 0.')

    def test_check_ingest_queues_db_error(self):
        request = self.getRequest()
        with patch.object(request, 'db', new_callable=self.makeFailCallable):
            check = check_ingest_queues(request)
        self.assertCheck(
            check, 'Ingest queue sizes', 'Fail',
            'Cannot query the ingest queues.')

    def test_check_bzr_revision(self):
        request = self.getRequest()
        check = check_bzr_revision(request)
        self.assertCheck(check, 'BZR revision', 'Pass')
        # check.remark is an integer: No exception is raised.
        int(check.remark)

    def test_check_bzr_revision_error(self):
        request = self.getRequest()
        from charmworld import health
        with patch.object(health, 'get_transport',
                          new_callable=self.makeFailCallable):
            check = check_bzr_revision(request)
        remark = "Can't retrieve revision number."
        self.assertCheck(check, 'BZR revision', 'Fail', remark)

    def test_check_elasticsearch_status(self):
        self.use_index_client()
        request = self.getRequest()
        check = check_elasticsearch_status(request)
        self.assertCheck(check, 'ElasticSearch server', 'Pass')
        self.assertTrue(check.remark.startswith('status: green, nodes:'))

    def test_check_elasticsearch_status_es_yellow_status(self):
        index_client = self.use_index_client()

        def make_health_method():
            def health():
                return {
                    'status': 'yellow',
                    'number_of_nodes': 1,
                    'active_primary_shards': 0,
                    'active_shards': 1,
                    'initializing_shards': 0,
                    'relocating_shards': 0,
                }
            return health

        with patch.object(index_client, 'health',
                          new_callable=make_health_method):
            request = self.getRequest()
            check = check_elasticsearch_status(request)
        self.assertCheck(check, 'ElasticSearch server', 'Pass')
        self.assertTrue(check.remark.startswith('status: yellow, nodes:'))

    def test_check_elasticsearch_status_es_red_status(self):
        index_client = self.use_index_client()

        def make_health_method():
            def health():
                return {
                    'status': 'red',
                    'number_of_nodes': 1,
                    'active_primary_shards': 0,
                    'active_shards': 1,
                    'initializing_shards': 0,
                    'relocating_shards': 0,
                }
            return health

        with patch.object(index_client, 'health',
                          new_callable=make_health_method):
            request = self.getRequest()
            check = check_elasticsearch_status(request)
        self.assertCheck(check, 'ElasticSearch server', 'Fail')
        self.assertTrue(check.remark.startswith('status: red, nodes:'))

    def test_check_elasticsearch_status_es_access_failing(self):
        index_client = self.use_index_client()
        with patch.object(index_client, 'health',
                          new_callable=self.makeFailCallable):
            request = self.getRequest()
            check = check_elasticsearch_status(request)
            remark = "Can't retrieve the ES server's health status."
        self.assertCheck(check, 'ElasticSearch server', 'Fail', remark)

    def test_check_crontab_pass(self):
        def make_fake_listdir():
            def fake_listdir(path):
                return ('foo', 'charmworld')
            return fake_listdir

        with patch.object(os, 'listdir', new_callable=make_fake_listdir):
            request = self.getRequest()
            check = check_crontab(request)
        remark = 'Crontab file for charmworld exists.'
        self.assertCheck(check, 'Crontab', 'Pass', remark)

    def test_check_crontab_fail(self):
        def make_fake_listdir():
            def fake_listdir(path):
                return ('foo', )
            return fake_listdir

        with patch.object(os, 'listdir', new_callable=make_fake_listdir):
            request = self.getRequest()
            check = check_crontab(request)
        remark = 'Crontab file for charmworld not found.'
        self.assertCheck(check, 'Crontab', 'Fail', remark)

    def test_check_server_start_time_pass(self):
        def make_uname():
            def uname():
                return (None, 'foo')
            return uname

        with patch.object(os, 'uname', new_callable=make_uname):
            record_server_start_time(self.db)
        check = check_server_start_time(self.getRequest())
        self.assertCheck(check, 'Server started', 'Pass')
        mo = re.search(
            '^foo: \d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\dZ$', check.remark)
        self.assertIsNot(None, mo)

    def test_check_server_start_time_fail(self):
        check = check_server_start_time(self.getRequest())
        self.assertCheck(
            check, 'Server started', 'Fail',
            "Can't retrieve the server's start time.")

    def test_check_last_ingest_job_pass(self):
        set_status_timestamp(self.db, LAST_INGEST_JOB_FINISHED)
        check = check_last_ingest_job(self.getRequest())
        self.assertCheck(check, 'Last ingest job', 'Pass')
        mo = re.search(
            '^finished at \d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\dZ$', check.remark)
        self.assertIsNot(None, mo)

    def test_check_last_ingest_job_fail(self):
        check = check_last_ingest_job(self.getRequest())
        self.assertCheck(
            check, 'Last ingest job', 'Fail',
            "Can't retrieve the time when the last ingest job finished.")
