# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from charmworld.models import (
    CharmSource,
)
from charmworld.search import (
    BUNDLE,
    CHARM,
)
from charmworld.testing import (
    factory,
    login_request,
    ViewTestBase,
    WebTestBase,
)
from charmworld.views.delete import (
    bundle_delete,
    charm_delete,
    delete_confirm,
)


class TestDeleteCharm(ViewTestBase):

    def test_form_renders(self):
        """Sanity check that the form renders without errors."""
        cid, charm_data = factory.makeCharm(
            self.db, promulgated=False, owner='foo')
        request = self.getRequest()
        request.matchdict = {
            'charm': charm_data['name'],
            'series': charm_data['series'],
        }
        response = delete_confirm(request, charm_data, CHARM, '')

        self.assertIn('form', response)
        self.assertIn('<label', response['form'])

    def test_form_value_unchecked(self):
        """GET requests result checkbox always being false."""
        cid, charm_data = factory.makeCharm(
            self.db, promulgated=False, owner='foo')
        request = self.getRequest()
        request.matchdict = {
            'charm': charm_data['name'],
            'series': charm_data['series'],
        }
        response = delete_confirm(request, charm_data, CHARM, '')
        self.assertFalse(response['values']['delete_it'])

    def test_canceling_does_not_delete_the_charm(self):
        """Verify canceling the form does not cause delete the charm."""
        charm_data = factory.get_charm_json(name='name1', promulgated=False)
        self.use_index_client()
        self.enable_routes()
        charm_source = CharmSource.from_request(self)
        charm_source.save(charm_data)
        # Add the submit button to trigger the right path through the view.
        post_data = {
            '_charset_': 'UTF-8',
            '__formid__': 'deform',
            'cancel': 'cancel',
            'delete_it': 'true',
        }
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'charm': charm_data['name'],
            'series': charm_data['series'],
        }
        # We can find the charm in the database.
        self.assertTrue(self.db.charms.find_one({'_id': charm_data['_id']}))
        # And after submitting the form with the checkbox unchecked, the charm
        # remains.
        delete_confirm(request, charm_data, CHARM, '/')
        self.assertTrue(self.db.charms.find_one({'_id': charm_data['_id']}))

    def test_not_checking_checkbox_does_not_delete_the_charm(self):
        """Verify not checking the checkbox does not cause delete the charm."""
        charm_data = factory.get_charm_json(name='name1', promulgated=False)
        self.use_index_client()
        self.enable_routes()
        charm_source = CharmSource.from_request(self)
        charm_source.save(charm_data)
        # Add the submit button to trigger the right path through the view.
        post_data = {
            '_charset_': 'UTF-8',
            '__formid__': 'deform',
            'submit': 'submit',
            'delete_it': 'false',
        }
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'charm': charm_data['name'],
            'series': charm_data['series'],
        }
        # We can find the charm in the database.
        self.assertTrue(self.db.charms.find_one({'_id': charm_data['_id']}))
        # And after submitting the form with the checkbox unchecked, the charm
        # remains.
        delete_confirm(request, charm_data, CHARM, '/')
        self.assertTrue(self.db.charms.find_one({'_id': charm_data['_id']}))

    def test_checking_checkbox_deletes_charm(self):
        """Verify checking the checkbox causes the charm to be deleted."""
        charm_data = factory.get_charm_json(name='name1', promulgated=False)
        self.use_index_client()
        self.enable_routes()
        charm_source = CharmSource.from_request(self)
        charm_source.save(charm_data)
        # Add the submit button to trigger the right path through the view.
        post_data = {
            '_charset_': 'UTF-8',
            '__formid__': 'deform',
            'submit': 'submit',
            'delete_it': 'true',
        }
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'charm': charm_data['name'],
            'series': charm_data['series'],
        }
        # We can find the charm in the database.
        self.assertTrue(self.db.charms.find_one({'_id': charm_data['_id']}))
        # But after deleting the charm is not found.
        delete_confirm(request, charm_data, CHARM, '/')
        self.assertFalse(self.db.charms.find_one({'_id': charm_data['_id']}))

    def test_redirect_on_submit(self):
        """After a submit, an HTTP redirect is raised with the given URL."""
        charm = factory.get_charm_json(name='name1', promulgated=False)
        self.use_index_client()
        self.enable_routes()
        charm_source = CharmSource.from_request(self)
        charm_source.save(charm)
        # Add the submit button to trigger the right path through the view.
        post_data = {
            '_charset_': 'UTF-8',
            '__formid__': 'deform',
            'submit': 'submit',
            'delete_it': 'false',
        }
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }

        response = delete_confirm(request, charm, 'charm', 'URL')
        self.assertEqual(302, response.status_code)
        self.assertEqual(response.headers['Location'], 'URL')


class TestDeleteBundle(ViewTestBase):

    def test_form_renders(self):
        """Sanity check that the form renders without errors."""
        bundle, bundle_data = factory.makeBundle(
            self.db, promulgated=False, owner='foo')
        request = self.getRequest()
        request.matchdict = {
            'bundle': bundle_data['name'],
            'basket': bundle_data['basket_name'],
        }
        response = delete_confirm(request, bundle_data, BUNDLE, '')

        self.assertIn('form', response)
        self.assertIn('<label', response['form'])

    def test_form_value_unchecked(self):
        """GET requests result checkbox always being false."""
        bundle, bundle_data = factory.makeBundle(
            self.db, promulgated=False, owner='foo')
        request = self.getRequest()
        request.matchdict = {
            'bundle': bundle_data['name'],
            'basket': bundle_data['basket_name'],
        }
        response = delete_confirm(request, bundle_data, BUNDLE, '')
        self.assertFalse(response['values']['delete_it'])

    def test_canceling_does_not_delete_the_bundle(self):
        """Verify canceling the form does not cause delete the bundle."""
        bundle_data = factory.get_bundle_data(name='name1', promulgated=False)
        self.use_index_client()
        self.enable_routes()
        self.db.bundles.save(bundle_data)
        # Add the submit button to trigger the right path through the view.
        post_data = {
            '_charset_': 'UTF-8',
            '__formid__': 'deform',
            'cancel': 'cancel',
            'delete_it': 'true',
        }
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'bundle': bundle_data['name'],
            'basket': bundle_data['basket_name'],
        }
        # We can find the bundle in the database.
        self.assertTrue(self.db.bundles.find_one({'_id': bundle_data['_id']}))
        # And after submitting the form with the checkbox unchecked, the bundle
        # remains.
        delete_confirm(request, bundle_data, BUNDLE, '/')
        self.assertTrue(self.db.bundles.find_one({'_id': bundle_data['_id']}))

    def test_not_checking_checkbox_does_not_delete_the_bundle(self):
        """Verify not checking the checkbox does not delete the bundle."""
        bundle_data = factory.get_bundle_data(name='name1', promulgated=False)
        self.use_index_client()
        self.enable_routes()
        self.db.bundles.save(bundle_data)
        # Add the submit button to trigger the right path through the view.
        post_data = {
            '_charset_': 'UTF-8',
            '__formid__': 'deform',
            'submit': 'submit',
            'delete_it': 'false',
        }
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'bundle': bundle_data['name'],
            'basket': bundle_data['basket_name'],
        }
        # We can find the bundle in the database.
        self.assertTrue(self.db.bundles.find_one({'_id': bundle_data['_id']}))
        # And after submitting the form with the checkbox unchecked, the bundle
        # remains.
        delete_confirm(request, bundle_data, BUNDLE, '/')
        self.assertTrue(self.db.bundles.find_one({'_id': bundle_data['_id']}))

    def test_checking_checkbox_deletes_bundle(self):
        """Verify checking the checkbox causes the bundle to be deleted."""
        bundle_data = factory.get_bundle_data(name='name1', promulgated=False)
        self.use_index_client()
        self.enable_routes()
        self.db.bundles.save(bundle_data)
        # Add the submit button to trigger the right path through the view.
        post_data = {
            '_charset_': 'UTF-8',
            '__formid__': 'deform',
            'submit': 'submit',
            'delete_it': 'true',
        }
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'bundle': bundle_data['name'],
            'basket': bundle_data['basket_name'],
        }
        # We can find the bundle in the database.
        query = {'_id': bundle_data['_id']}
        self.assertIsNotNone(
            self.db.bundles.find_one(query))
        # But after deleting the bundle is not found.
        delete_confirm(request, bundle_data, BUNDLE, '/')
        self.assertIsNone(
            self.db.bundles.find_one(query))

    def test_redirect_on_submit(self):
        """After a submit, an HTTP redirect is raised with the given URL."""
        bundle_data = factory.get_bundle_data(name='name1', promulgated=False)
        self.use_index_client()
        self.enable_routes()
        self.db.bundles.save(bundle_data)
        # Add the submit button to trigger the right path through the view.
        post_data = {
            '_charset_': 'UTF-8',
            '__formid__': 'deform',
            'submit': 'submit',
            'delete_it': 'false',
        }
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'bundle': bundle_data['name'],
            'basket': bundle_data['basket_name'],
        }

        response = delete_confirm(request, bundle_data, 'bundle', 'URL')
        self.assertEqual(302, response.status_code)
        self.assertEqual(response.headers['Location'], 'URL')


class TestDeleteCharmForm(ViewTestBase, WebTestBase):

    def setUp(self):
        super(TestDeleteCharmForm, self).setUp()
        # Register the routes so URL construction will work in the tests.
        self.enable_routes()

    def test_form_renders(self):
        """Sanity check that the form renders without errors."""
        cid, charm = factory.makeCharm(self.db, promulgated=True, owner='foo')
        request = self.getRequest()
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }
        response = charm_delete(request)

        self.assertIn('form', response)
        self.assertIn('<label', response['form'])

    def test_charm_delete_route(self):
        """The charm delete form is wired up to a route."""
        charm = factory.makeCharm(self.db, promulgated=True, owner='foo')[1]
        url = '/charms/{series}/{name}/delete'.format(**charm)
        with login_request(self.app, groups=['charmers']):
            # Get the URL and assert the given status code.
            self.app.get(url, status=200)

    def test_personal_charm_delete_route(self):
        """The personal charm delete form is wired up to a route."""
        charm = factory.makeCharm(self.db, promulgated=False, owner='foo')[1]
        url = '/~{owner}/{series}/{name}/delete'.format(**charm)
        with login_request(self.app, groups=['charmers']):
            # Get the URL and assert the given status code.
            self.app.get(url, status=200)

    def test_personal_charm_rev_delete_route(self):
        """The personal charm delete with revision has no route."""
        charm = factory.makeCharm(self.db, promulgated=False, owner='foo')[1]
        url = '/~{owner}/{series}/{name}-{revision}/delete'.format(**charm)
        with login_request(self.app, groups=['charmers']):
            # Get the URL and assert the given status code.
            self.app.get(url, status=404)


class TestDeleteBundleForm(ViewTestBase, WebTestBase):

    def setUp(self):
        super(TestDeleteBundleForm, self).setUp()
        # Register the routes so URL construction will work in the tests.
        self.enable_routes()

    def test_form_renders(self):
        """Sanity check that the form renders without errors."""
        bundle, bundle_data = factory.makeBundle(self.db, promulgated=True)
        request = self.getRequest()
        request.matchdict = {
            'bundle': bundle_data['name'],
            'basket': bundle_data['basket_name'],
        }
        response = bundle_delete(request)

        self.assertIn('form', response)
        self.assertIn('<label', response['form'])

    def test_bundle_delete_route(self):
        """The bundle delete form is wired up to a route."""
        bundle, bundle_data = factory.makeBundle(self.db, promulgated=True)
        url = '/bundle/{basket_name}/{name}/delete'.format(
            **bundle_data)
        with login_request(self.app, groups=['charmers']):
            # Get the URL and assert the given status code.
            self.app.get(url, status=200)

    def test_personal_bundle_delete_route(self):
        """The personal bundle delete form is wired up to a route."""
        bundle, bundle_data = factory.makeBundle(self.db, promulgated=False)
        url = '/bundle/~{owner}/{basket_name}/{name}/delete'.format(
            **bundle_data)
        with login_request(self.app, groups=['charmers']):
            # Get the URL and assert the given status code.
            self.app.get(url, status=200)


class TestCharmDeleteFormAuthorization(WebTestBase):

    def test_form_403s_for_non_charmers_promulgated(self):
        """The form is not permitted if you're not a charmer."""
        with login_request(self.app, groups=[]) as resp:
            cid, charm = factory.makeCharm(self.db, promulgated=True)
            # The status force the check against the status code.
            url = '/charms/{series}/{name}/delete'.format(**charm)
            resp = resp.goto(url, status=403)

    def test_form_403s_for_non_charmers_non_promulgated(self):
        """The form is not permitted if you're not a charmer."""
        with login_request(self.app, groups=[]) as resp:
            cid, charm = factory.makeCharm(self.db, promulgated=False)
            # The status force the check against the status code.
            url = '/~{owner}/{series}/{name}/delete'.format(**charm)
            resp = resp.goto(url, status=403)

    def test_form_secured_from_users_promulgated(self):
        """The form is only allowed to be edited/submitted by charmers."""
        with login_request(self.app, groups=['charmers']) as resp:
            cid, charm = factory.makeCharm(self.db, promulgated=True)
            url = '/charms/{series}/{name}/delete'.format(**charm)
            resp = resp.goto(url, status=200)

    def test_form_secured_from_users_non_promulgated(self):
        """The form is only allowed to be edited/submitted by charmers."""
        with login_request(self.app, groups=['charmers']) as resp:
            cid, charm = factory.makeCharm(
                self.db, promulgated=False, owner='benji')
            url = '/~{owner}/{series}/{name}/delete'.format(**charm)
            resp = resp.goto(url, status=200)


class TestBundleDeleteFormAuthorization(WebTestBase):

    def test_form_403s_for_non_charmers_promulgated(self):
        """The form is not permitted if you're not a charmer."""
        with login_request(self.app, groups=[]) as resp:
            bundle, bundle_data = factory.makeBundle(self.db, promulgated=True)
            # The status force the check against the status code.
            url = '/bundle/{basket_name}/{name}/delete'.format(
                **bundle_data)
            resp = resp.goto(url, status=403)

    def test_form_403s_for_non_charmers_non_promulgated(self):
        """The form is not permitted if you're not a charmer."""
        with login_request(self.app, groups=[]) as resp:
            bundle, bundle_data = factory.makeBundle(
                self.db, promulgated=False)
            # The status force the check against the status code.
            url = '/bundle/~{owner}/{basket_name}/{name}/delete'.format(
                **bundle_data)
            resp = resp.goto(url, status=403)

    def test_form_secured_from_users_promulgated(self):
        """The form is only allowed to be edited/submitted by charmers."""
        with login_request(self.app, groups=['charmers']) as resp:
            bundle, bundle_data = factory.makeBundle(self.db, promulgated=True)
            url = '/bundle/{basket_name}/{name}/delete'.format(
                **bundle_data)
            resp = resp.goto(url, status=200)

    def test_form_secured_from_users(self):
        """The form is only allowed to be edited/submitted by charmers."""
        with login_request(self.app, groups=['charmers']) as resp:
            bundle, bundle_data = factory.makeBundle(
                self.db, promulgated=False, owner='benji')
            url = '/bundle/~{owner}/{basket_name}/{name}/delete'.format(
                **bundle_data)
            resp = resp.goto(url, status=200)
