from textwrap import dedent
import yaml

from charmworld.models import Charm
from charmworld.testing import (
    factory,
    load_data_file,
)


class TestBundleProof:
    """Verify that a bundle can be proofed in charmworld."""

    def setup_charm(self, description=''):
        # Create a charm for use in the test.  Return it as a Charm object.
        _id, charm = factory.makeCharm(
            self.db,
            description='',
        )
        return Charm(charm)

    def setup_relatable_charms(self, second_requires, second_provides):
        # Unless otherwise specified, the generated charm provides
        # website:https and requires database:mongodb.
        _id, charm_one = factory.makeCharm(
            self.db,
            description='',
            store_revision=0,
        )

        _id, charm_two = factory.makeCharm(
            self.db,
            description='',
            requires=second_requires,
            provides=second_provides,
        )

        charm_one = Charm(charm_one)
        one_store_url = charm_one.store_url

        charm_two = Charm(charm_two)
        two_store_url = charm_two.store_url

        return one_store_url, two_store_url

    def test_bundle_proof_no_deployer_file(self):
        request = self.make_request('bundle', remainder='/proof')
        response = self.api_class(request)()
        self.assertEqual(400, response.status_code)
        results = response.json_body
        self.assertEqual(0, len(results['warning_messages']))
        self.assertEqual(
            ['No deployer file data received.'],
            results['error_messages'])

    def test_bundle_proof_invalid_deployer_file(self):
        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': '{',
        }
        response = self.api_class(request)()
        self.assertEqual(400, response.status_code)
        results = response.json_body
        self.assertEqual(1, len(results['error_messages']))
        self.assertEqual(0, len(results['warning_messages']))
        self.assertIn(
            'Could not parse', results['error_messages'][0])

    def test_bundle_proof_charm_exists(self):
        deployer_file = load_data_file('sample_bundle/bundles.yaml')

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': deployer_file,
        }

        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body
        self.assertEqual(1, len(results['errors']))
        self.assertEqual(0, len(results['warning_messages']))
        self.assertEqual(4, len(results['errors'][0]['services'].keys()))
        self.assertIn(
            'Could not find ',
            results['errors'][0]['services']['db'][0]['message'])

        # And each failure has a list of 'process' information.
        self.assertIn(
            'debug_info',
            results['errors'][0]['services']['db'][0].keys())

    def test_bundle_proof_valid(self):
        charm = self.setup_charm()

        bundle_config = {
            'wiki': {
                'services': {
                    'charm': {
                        'charm_url': str(charm.store_url),
                    }
                }
            }
        }

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': yaml.dump(bundle_config),
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body
        self.assertEqual(0, len(results['errors']))
        self.assertEqual(1, len(results['services_info'].items()))
        self.assertEqual(charm.store_url,
                         results['services_info']['charm']['store_url'])

    def test_bundle_proof_invalid_config_type(self):
        charm = self.setup_charm()

        bundle_config = {
            'wiki': {
                'services': {
                    'charm': {
                        'charm_url': str(charm.store_url),
                        'options': {
                            'script-interval': 'invalid',
                        }
                    }
                }
            }
        }

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': yaml.dump(bundle_config),
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body
        self.assertEqual(1, len(results['errors']))
        self.assertEqual(1, len(results['error_messages']))
        self.assertEqual(0, len(results['warning_messages']))

        # Note that the View will prefix the error message from the proof with
        # the bundle name to help identify where the error message came from.
        self.assertEqual(
            ['wiki/charm: script-interval is not of type int.'],
            results['error_messages'])

    def test_bundle_proof_supports_multiple_errors(self):
        charm = self.setup_charm()

        bundle_config = {
            'wiki': {
                'services': {
                    'charm': {
                        'charm_url': str(charm.store_url),
                        'options': {
                            'script-interval': 'invalid',
                            'no-exist': 'hah invalid',
                        }
                    },
                    'fail': {
                        'name': 'will-fail'
                    }
                }
            }
        }

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': yaml.dump(bundle_config),
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body
        self.assertEqual(1, len(results['errors']))
        self.assertEqual(3, len(results['error_messages']))
        self.assertEqual(0, len(results['warning_messages']))

        # One message from not finding a charm.
        self.assertEqual(
            ['wiki/fail: Could not find charm for service.',
             'wiki/charm: script-interval is not of type int.',
             'wiki/charm: The charm has no option for: no-exist'],
            results['error_messages'])

    def test_bundle_proof_errors_on_invalid_relations(self):
        # The default charm provides website:https and requires
        # database:mongodb
        one_store_url, two_store_url = self.setup_relatable_charms(
            {
                'proxy': {
                    'interface': 'http',
                }
            },
            {
                'cache': {
                    'interface': 'memcache',
                },
                'db': {
                    'interface': 'mongodb',
                }
            },
        )

        bundle_base = dedent("""\
            wiki:
              series: precise
              promulgated: True
              services:
                wiki:
                  charm: {}
                db:
                  charm: {}
              relations:
                - ["wiki:proxy", db]""").format(one_store_url, two_store_url)

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': bundle_base,
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body
        self.assertEqual(1, len(results['error_messages']))
        self.assertEqual(0, len(results['warning_messages']))
        self.assertIn(
            'wiki: Invalid relation requested: proxy',
            response.json_body['error_messages'][0])

    def test_bundle_proof_errors_on_incompatible_relations(self):
        one_store_url, two_store_url = self.setup_relatable_charms(
            {
                'proxy': {
                    'interface': 'http',
                }
            }, {
                'cache': {
                    'interface': 'memcache'
                },
                'db': {
                    'interface': 'mongodb'
                }
            },
        )

        bundle_base = dedent("""\
            wiki:
              series: precise
              promulgated: True
              services:
                wiki:
                  charm: {}
                db:
                  charm: {}
              relations:
                - ["wiki:db", "db:proxy"]""").format(
            one_store_url, two_store_url)

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': bundle_base,
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body
        self.assertEqual(1, len(results['error_messages']))
        self.assertEqual(0, len(results['warning_messages']))
        self.assertIn(
            'wiki: The requested relation db to proxy is incompatible',
            response.json_body['error_messages'][0])

    def test_bundle_proof_requires_both_endpoints_are_compatible(self):
        # If one endpoint provides "proxy" but the other doesn't, then a
        # relation from "proxy" to "proxy" should not work.
        one_store_url, two_store_url = self.setup_relatable_charms(
            {
                'proxy': {
                    'interface': 'http',
                }
            }, {
                'cache': {
                    'interface': 'memcache'
                },
                'db': {
                    'interface': 'mongodb'
                }
            },
        )

        bundle_base = dedent("""\
            wiki:
              series: precise
              promulgated: True
              services:
                wiki:
                  charm: {}
                db:
                  charm: {}
              relations:
                - ["wiki:proxy", "db:proxy"]""").format(
            one_store_url, two_store_url)

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': bundle_base,
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body
        self.assertEqual(1, len(results['error_messages']))
        self.assertEqual(0, len(results['warning_messages']))
        self.assertIn(
            'wiki: The requested relation proxy to proxy is incompatible',
            response.json_body['error_messages'][0])

    def test_bundle_proof_passes_with_valid_relations(self):
        one_store_url, two_store_url = self.setup_relatable_charms(
            {
                'proxy': {
                    'interface': 'https',
                }
            }, {
                'cache': {
                    'interface': 'memcache'
                },
                'db': {
                    'interface': 'mongodb'
                }
            }
        )

        bundle_base = dedent("""\
            wiki:
              series: precise
              promulgated: True
              services:
                wiki:
                  charm: {}
                db:
                  charm: {}
              relations:
                - ["wiki:database", "db:db"]""").format(
            one_store_url, two_store_url)

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': bundle_base,
        }
        response = self.api_class(request)()
        results = response.json_body
        self.assertEqual(0, len(results['error_messages']))
        self.assertEqual(0, len(results['warning_messages']))
        self.assertEqual(200, response.status_code)

    def test_bundle_proof_passes_with_reversed_relations(self):
        one_store_url, two_store_url = self.setup_relatable_charms(
            {
                'proxy': {
                    'interface': 'https',
                }
            }, {
                'cache': {
                    'interface': 'memcache'
                },
                'db': {
                    'interface': 'mongodb'
                }
            }
        )

        bundle_base = dedent("""\
            wiki:
              series: precise
              promulgated: True
              services:
                wiki:
                  charm: {}
                db:
                  charm: {}
              relations:
                - ["db:db", "wiki:database"]""").format(
            one_store_url, two_store_url)

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': bundle_base,
        }
        response = self.api_class(request)()
        results = response.json_body
        self.assertEqual([], results['error_messages'])
        self.assertEqual([], results['warning_messages'])
        self.assertEqual(200, response.status_code)

    def test_bundle_proof_passes_with_valid_vague_as_can_be(self):
        one_store_url, two_store_url = self.setup_relatable_charms(
            {
                'proxy': {
                    'interface': 'https',
                }
            }, {
                'cache': {
                    'interface': 'memcache'
                },
                'db': {
                    'interface': 'mongodb'
                }
            },
        )

        bundle_base = dedent("""
            wiki:
              series: precise
              promulgated: True
              services:
                wiki:
                  charm: {}
                db:
                  charm: {}
              relations:
                - ["wiki", "db"]""").format(one_store_url, two_store_url)

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': bundle_base,
        }
        response = self.api_class(request)()
        results = response.json_body
        self.assertEqual(0, len(results['error_messages']))
        self.assertEqual(0, len(results['warning_messages']))
        self.assertEqual(200, response.status_code)

    def test_bundle_proof_errors_with_missing_service(self):
        one_store_url, two_store_url = self.setup_relatable_charms(
            {
                'proxy': {
                    'interface': 'https',
                }
            }, {
                'cache': {
                    'interface': 'memcache'
                },
                'db': {
                    'interface': 'mongodb'
                }
            },
        )

        bundle_base = dedent("""
            wiki:
              series: precise
              promulgated: True
              services:
                wiki:
                  charm: {}
                db:
                  charm: {}
              relations:
                - ["wiki", "party"]""").format(one_store_url, two_store_url)

        request = self.make_request('bundle', remainder='/proof')
        request.params = {'deployer_file': bundle_base}
        response = self.api_class(request)()
        results = response.json_body
        self.assertEqual(1, len(results['error_messages']))
        self.assertEqual(0, len(results['warning_messages']))
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            'wiki: Invalid relation to charm: party not found.',
            results['error_messages'][0])

    def test_bundle_proof_with_inheritance(self):
        # Show that inheritance works by relating a charm in wiki-ext to one
        # found only in the inherited bundle.
        url1, url2 = self.setup_relatable_charms(
            {
                'proxy': {
                    'interface': 'https',
                }
            }, {
                'cache': {
                    'interface': 'memcache'
                },
                'db': {
                    'interface': 'mongodb'
                }
            },
        )

        requires = {
            'juju-info': {
                'interface': 'juju-info',
                'scope': 'container',
            }
        }
        _id, charm3 = factory.makeCharm(
            self.db,
            description='logger',
            store_revision=0,
            requires=requires,
        )
        charm3 = Charm(charm3)
        url3 = charm3.store_url

        bundle_base = dedent("""
            wiki:
              series: precise
              promulgated: True
              services:
                wiki:
                  charm: {}
                db:
                  charm: {}
              relations:
                - ["wiki", "db"]

            wiki-ext:
              inherits: wiki
              services:
                log:
                  charm: {}
              relations:
                - ["wiki", "log"]
                - ["log", "db"]""").format(url1, url2, url3)

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': bundle_base,
        }
        response = self.api_class(request)()
        results = response.json_body
        self.assertEqual(0, len(results['error_messages']))
        self.assertEqual(0, len(results['warning_messages']))
        self.assertEqual(200, response.status_code)

    def test_bundle_proof_invalid_constraints(self):
        charm = self.setup_charm()

        bundle_config = {
            'wiki': {
                'services': {
                    'charm': {
                        'charm_url': str(charm.store_url),
                        'constraints': 'cows=1 cpu-donkeys=4',
                    }
                }
            }
        }

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': yaml.dump(bundle_config),
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body
        self.assertEqual(1, len(results['errors']))
        self.assertEqual(1, len(results['error_messages']))
        self.assertEqual(0, len(results['warning_messages']))
        self.assertEqual(
            ['wiki/charm: unsupported constraints: cows, cpu-donkeys'],
            results['error_messages'])

    def test_bundle_proof_comma_separated_constraints(self):
        charm = self.setup_charm()

        bundle_config = {
            'wiki': {
                'services': {
                    'charm': {
                        'charm_url': str(charm.store_url),
                        'constraints': 'mem=1G,cpu-cores=4',
                    }
                }
            }
        }

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': yaml.dump(bundle_config),
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body
        self.assertEqual(1, len(results['errors']))
        self.assertEqual(0, len(results['error_messages']))
        self.assertEqual(1, len(results['warning_messages']))
        expected = ("wiki/charm: The constraints are comma-separated, "
                    "which is deprecated.")
        self.assertEqual([expected], results['warning_messages'])

    def test_bundle_proof_multi_error_constraints(self):
        # If there is a failure due to invalid constraints and comma-separated,
        # we only report the former.
        charm = self.setup_charm()

        bundle_config = {
            'wiki': {
                'services': {
                    'charm': {
                        'charm_url': str(charm.store_url),
                        'constraints': 'mem=1G,cpu-donkeys=4',
                    }
                }
            }
        }

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': yaml.dump(bundle_config),
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body

        self.assertEqual(1, len(results['errors']))
        self.assertEqual(1, len(results['error_messages']))
        self.assertEqual(0, len(results['warning_messages']))

        expected = 'wiki/charm: unsupported constraints: cpu-donkeys'
        self.assertEqual([expected], results['error_messages'])

    def test_bundle_with_juju_gui(self):
        # If a service with the name "juju-gui" is part of the bundle then
        # proof should error.
        charm = self.setup_charm()
        bundle_config = {
            'wiki': {
                'services': {
                    'juju-gui': {
                        'charm_url': str(charm.store_url),
                    }
                }
            }
        }
        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': yaml.dump(bundle_config),
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body
        self.assertEqual(1, len(results['errors']))
        self.assertEqual(1, len(results['error_messages']))
        self.assertEqual(0, len(results['warning_messages']))
        expected = (
            'wiki/juju-gui: Bundles may not have services named "juju-gui".')
        self.assertEqual([expected], results['error_messages'])
